/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#include "gdl-dock.h"
#include "gdl-dock-item.h"
#include "gdl-dock-paned.h"

#define GDL_DOCK_ITEM_DEBUG

#if defined GDL_DOCK_ITEM_DEBUG && defined __GNUC__
#define DEBUG(x)                                        \
  do                                                    \
    {                                                   \
      g_print ("%s.%d: ", __FUNCTION__, __LINE__);      \
      g_print x;                                        \
      g_print ("\n");                                   \
    }                                                   \
  while (0)
#else
#define DEBUG(x)
#endif


/* Private prototypes */

static void  gdl_dock_item_class_init    (GdlDockItemClass *class);
static void  gdl_dock_item_init          (GdlDockItem *dock_item);

static void  gdl_dock_item_set_arg       (GtkObject *object,
                                          GtkArg *arg,
                                          guint arg_id);
static void  gdl_dock_item_get_arg       (GtkObject *object,
                                          GtkArg *arg,
                                          guint arg_id);
static void  gdl_dock_item_destroy       (GtkObject *object);

static void  gdl_dock_item_size_request  (GtkWidget *widget,
                                          GtkRequisition *requisition);
static void  gdl_dock_item_size_allocate (GtkWidget *widget,
                                          GtkAllocation *allocation);
static void  gdl_dock_item_map           (GtkWidget *widget);
static void  gdl_dock_item_unmap         (GtkWidget *widget);
static void  gdl_dock_item_draw          (GtkWidget *widget,
                                          GdkRectangle *area);
static gint  gdl_dock_item_expose        (GtkWidget *widget,
                                          GdkEventExpose *event);
static void  gdl_dock_item_realize       (GtkWidget *widget);
static void  gdl_dock_item_unrealize     (GtkWidget *widget);
static void  gdl_dock_item_style_set     (GtkWidget *widget,
                                          GtkStyle  *previous_style);
static void  gdl_dock_item_set_floating  (GdlDockItem *item, 
                                          gboolean val);
static gint  gdl_dock_item_button_changed (GtkWidget *widget,
                                           GdkEventButton *event);
static gint  gdl_dock_item_delete_event  (GtkWidget *widget,
                                          GdkEventAny *event);
static gint  gdl_dock_item_motion        (GtkWidget *widget,
                                          GdkEventMotion *event);
static void  gdl_dock_item_grab_pointer  (GdlDockItem *item);


/* Class variables and definitions */

enum {
    ARG_0,
    ARG_ORIENTATION,
    ARG_RESIZE,
    ARG_SHRINK
};

static GtkBinClass *parent_class = NULL;

/* FIXME: should be an arg of the widget */
#define DRAG_HANDLE_SIZE  10
#define GDL_DOCK_ITEM_NOT_LOCKED(x) (!(GDL_DOCK_ITEM (x)->behavior && GDL_DOCK_ITEM_BEH_LOCKED))

/* FIXME: implement the rest of the behaviors */


/* Private functions */

static void
gdl_dock_item_class_init (GdlDockItemClass *klass)
{
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;
    GtkContainerClass *container_class;

    object_class = (GtkObjectClass *) klass;
    widget_class = (GtkWidgetClass *) klass;
    container_class = (GtkContainerClass *) klass;

    parent_class = gtk_type_class (gtk_bin_get_type ());

    gtk_object_add_arg_type ("GdlDockItem::orientation",
                             GTK_TYPE_ORIENTATION, GTK_ARG_READWRITE,
                             ARG_ORIENTATION);

    gtk_object_add_arg_type ("GdlDockItem::resize",
                             GTK_TYPE_BOOL, GTK_ARG_READWRITE,
                             ARG_RESIZE);

    gtk_object_add_arg_type ("GdlDockItem::shrink",
                             GTK_TYPE_BOOL, GTK_ARG_READWRITE,
                             ARG_SHRINK);

    object_class->set_arg = gdl_dock_item_set_arg;
    object_class->get_arg = gdl_dock_item_get_arg;
    object_class->destroy = gdl_dock_item_destroy;

    widget_class->realize = gdl_dock_item_realize;
    widget_class->unrealize = gdl_dock_item_unrealize;
    widget_class->map = gdl_dock_item_map;
    widget_class->unmap = gdl_dock_item_unmap;
    widget_class->size_request = gdl_dock_item_size_request;
    widget_class->size_allocate = gdl_dock_item_size_allocate;
    widget_class->style_set = gdl_dock_item_style_set;
    widget_class->draw = gdl_dock_item_draw;
    widget_class->expose_event = gdl_dock_item_expose;
    widget_class->button_press_event = gdl_dock_item_button_changed;
    widget_class->button_release_event = gdl_dock_item_button_changed;
    widget_class->motion_notify_event = gdl_dock_item_motion;
    widget_class->delete_event = gdl_dock_item_delete_event;

    klass->auto_reduce = NULL;
    klass->drag_request = NULL;
}

static void
gdl_dock_item_init (GdlDockItem *dock_item)
{
    GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (dock_item), GTK_NO_WINDOW);

    dock_item->name = NULL;
    dock_item->bin_window = NULL;
    dock_item->float_window = NULL;

    dock_item->is_floating = FALSE;
    dock_item->float_window_mapped = FALSE;
    dock_item->resize = TRUE;
    dock_item->shrink = FALSE;
    dock_item->orientation = GTK_ORIENTATION_HORIZONTAL;
    dock_item->behavior = GDL_DOCK_ITEM_BEH_NORMAL;

    dock_item->float_x = dock_item->float_y = 0;
    dock_item->dragoff_x = dock_item->dragoff_y = 0;
    dock_item->in_drag = FALSE;
}

static void
gdl_dock_item_set_arg (GtkObject *object,
                       GtkArg *arg,
                       guint arg_id)
{
    GdlDockItem *dock_item;

    dock_item = GDL_DOCK_ITEM (object);

    switch (arg_id) {
    case ARG_ORIENTATION:
        gdl_dock_item_set_orientation (dock_item, GTK_VALUE_ENUM (*arg));
        break;
    case ARG_RESIZE:
        dock_item->resize = GTK_VALUE_BOOL (*arg);
        gtk_widget_queue_resize (GTK_WIDGET (object));
        break;
    case ARG_SHRINK:
        dock_item->shrink = GTK_VALUE_BOOL (*arg);
        gtk_widget_queue_resize (GTK_WIDGET (object));
        break;
    default:
        break;
    }
}

static void
gdl_dock_item_get_arg (GtkObject *object,
                       GtkArg *arg,
                       guint arg_id)
{
    GdlDockItem *dock_item;

    dock_item = GDL_DOCK_ITEM (object);
    
    switch (arg_id) {
    case ARG_ORIENTATION:
        GTK_VALUE_ENUM (*arg) = dock_item->orientation;
        break;
    case ARG_RESIZE:
        GTK_VALUE_BOOL (*arg)= dock_item->resize;
        break;
    case ARG_SHRINK:
        GTK_VALUE_BOOL (*arg)= dock_item->shrink;
        break;
    default:
        arg->type = GTK_TYPE_INVALID;
        break;
    }
}

static void
gdl_dock_item_destroy (GtkObject *object)
{
    g_free (GDL_DOCK_ITEM (object)->name);

    if (GTK_OBJECT_CLASS (parent_class)->destroy)
        (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gdl_dock_item_size_request (GtkWidget *widget,
                            GtkRequisition *requisition)
{
    GtkBin *bin;
    GtkRequisition child_requisition;
    GdlDockItem *dock_item;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));
    g_return_if_fail (requisition != NULL);

    bin = GTK_BIN (widget);
    dock_item = GDL_DOCK_ITEM (widget);

    /* If our child is not visible, we still request its size, since
       we won't have any useful hint for our size otherwise.  */
    if (bin->child != NULL)
        gtk_widget_size_request (bin->child, &child_requisition);
    else {
        child_requisition.width = 0;
        child_requisition.height = 0;
    }

    if (dock_item->orientation == GTK_ORIENTATION_HORIZONTAL) {
        requisition->width = 
            GDL_DOCK_ITEM_NOT_LOCKED (dock_item) ? DRAG_HANDLE_SIZE : 0;
        if (bin->child != NULL) {
            requisition->width += child_requisition.width;
            requisition->height = child_requisition.height;
        } else
            requisition->height = 0;
    } else {
        requisition->height = 
            GDL_DOCK_ITEM_NOT_LOCKED (dock_item) ? DRAG_HANDLE_SIZE : 0;
        if (bin->child != NULL) {
            requisition->width = child_requisition.width;
            requisition->height += child_requisition.height;
        } else
            requisition->width = 0;
    }

    requisition->width += GTK_CONTAINER (widget)->border_width * 2;
    requisition->height += GTK_CONTAINER (widget)->border_width * 2;

    widget->requisition = *requisition;
}

static void
gdl_dock_item_size_allocate (GtkWidget *widget,
                             GtkAllocation *allocation)
{
    GtkBin *bin;
    GdlDockItem *dock_item;
  
    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));
    g_return_if_fail (allocation != NULL);
  
    DEBUG (("entering function"));

    bin = GTK_BIN (widget);
    dock_item = GDL_DOCK_ITEM (widget);

    widget->allocation = *allocation;

    DEBUG (("%s: %d %d %d %d", dock_item->name, 
            allocation->x, allocation->y, 
            allocation->width, allocation->height));

    if (GTK_WIDGET_REALIZED (widget))
        gdk_window_move_resize (widget->window,
                                widget->allocation.x,
                                widget->allocation.y,
                                widget->allocation.width,
                                widget->allocation.height);

    if (bin->child && GTK_WIDGET_VISIBLE (bin->child)) {
        GtkWidget *child;
        GtkAllocation child_allocation;
        int border_width;

        child = bin->child;
        border_width = GTK_CONTAINER (widget)->border_width;

        child_allocation.x = border_width;
        child_allocation.y = border_width;

        if (GDL_DOCK_ITEM_NOT_LOCKED (dock_item)) {
            if (dock_item->orientation == GTK_ORIENTATION_HORIZONTAL)
                child_allocation.x += DRAG_HANDLE_SIZE;
            else
                child_allocation.y += DRAG_HANDLE_SIZE;
        };

        if (dock_item->is_floating) {
            GtkRequisition child_requisition;
            guint float_width;
            guint float_height;

            gtk_widget_get_child_requisition (child, &child_requisition);

            child_allocation.width = child_requisition.width;
            child_allocation.height = child_requisition.height;

            float_width = child_allocation.width + 2 * border_width;
            float_height = child_allocation.height + 2 * border_width;
          
            if (dock_item->orientation == GTK_ORIENTATION_HORIZONTAL)
                float_width += DRAG_HANDLE_SIZE;
            else
                float_height += DRAG_HANDLE_SIZE;

            if (GTK_WIDGET_REALIZED (dock_item)) {
                gdk_window_resize (dock_item->float_window,
                                   float_width,
                                   float_height);
                gdk_window_move_resize (dock_item->bin_window,
                                        0,
                                        0,
                                        float_width,
                                        float_height);
            }
        } else {
            child_allocation.width = 
                MAX (1, (int) widget->allocation.width - 2 * border_width);
            child_allocation.height = 
                MAX (1, (int) widget->allocation.height - 2 * border_width);

            if (GDL_DOCK_ITEM_NOT_LOCKED (dock_item)) {
                if (dock_item->orientation == GTK_ORIENTATION_HORIZONTAL)
                    child_allocation.width = MAX 
                        ((int) child_allocation.width - DRAG_HANDLE_SIZE, 1);
                else
                    child_allocation.height = MAX 
                        ((int) child_allocation.height - DRAG_HANDLE_SIZE, 1);
            };

            if (GTK_WIDGET_REALIZED (dock_item))
                gdk_window_move_resize (dock_item->bin_window,
                                        0, 0,
                                        widget->allocation.width,
                                        widget->allocation.height);
        }

        gtk_widget_size_allocate (bin->child, &child_allocation);
    }
}

static void
gdl_dock_item_map (GtkWidget *widget)
{
    GtkBin *bin;
    GdlDockItem *di;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));

    GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

    bin = GTK_BIN (widget);
    di = GDL_DOCK_ITEM (widget);

    gdk_window_show (di->bin_window);
    if (! di->is_floating)
        gdk_window_show (widget->window);

    if (di->is_floating && !di->float_window_mapped)
        gdl_dock_item_window_float (di);

    if (bin->child
        && GTK_WIDGET_VISIBLE (bin->child)
        && !GTK_WIDGET_MAPPED (bin->child))
        gtk_widget_map (bin->child);
}

static void
gdl_dock_item_unmap (GtkWidget *widget)
{
    GdlDockItem *di;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));

    GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);

    di = GDL_DOCK_ITEM (widget);

    gdk_window_hide (widget->window);
    if (di->float_window_mapped) {
        gdk_window_hide (di->float_window);
        di->float_window_mapped = FALSE;
    }
}

static void
gdl_dock_item_realize (GtkWidget *widget)
{
    GdkWindowAttr attributes;
    gint attributes_mask;
    GdlDockItem *di;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));

    di = GDL_DOCK_ITEM (widget);

    GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual (widget);
    attributes.colormap = gtk_widget_get_colormap (widget);
    attributes.event_mask = (gtk_widget_get_events (widget)
                             | GDK_EXPOSURE_MASK);
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), 
                                     &attributes, attributes_mask);
    gdk_window_set_user_data (widget->window, widget);

    attributes.x = 0;
    attributes.y = 0;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.event_mask |= (gtk_widget_get_events (widget) |
                              GDK_EXPOSURE_MASK |
                              GDK_BUTTON1_MOTION_MASK |
                              GDK_POINTER_MOTION_HINT_MASK |
                              GDK_BUTTON_PRESS_MASK |
                              GDK_BUTTON_RELEASE_MASK);
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    di->bin_window = gdk_window_new (widget->window, &attributes, 
                                     attributes_mask);
    gdk_window_set_user_data (di->bin_window, widget);
    if (GTK_BIN (di)->child)
        gtk_widget_set_parent_window (GTK_BIN (di)->child, di->bin_window);

    attributes.x = 0;
    attributes.y = 0;
    attributes.width = widget->requisition.width;
    attributes.height = widget->requisition.height;
    attributes.window_type = GDK_WINDOW_TOPLEVEL;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual (widget);
    attributes.colormap = gtk_widget_get_colormap (widget);
    attributes.event_mask = (gtk_widget_get_events (widget) |
                             GDK_KEY_PRESS_MASK |
                             GDK_ENTER_NOTIFY_MASK |
                             GDK_LEAVE_NOTIFY_MASK |
                             GDK_FOCUS_CHANGE_MASK |
                             GDK_STRUCTURE_MASK);
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    di->float_window = gdk_window_new (NULL, &attributes, attributes_mask);
    gdk_window_set_transient_for (di->float_window, 
                                  gdk_window_get_toplevel(widget->window));
    gdk_window_set_user_data (di->float_window, widget);
    gdk_window_set_decorations (di->float_window, 0);
  
    widget->style = gtk_style_attach (widget->style, widget->window);
    gtk_style_set_background (widget->style, widget->window, 
                              GTK_WIDGET_STATE (di));
    gtk_style_set_background (widget->style, di->bin_window, 
                              GTK_WIDGET_STATE (di));
    gtk_style_set_background (widget->style, di->float_window, 
                              GTK_WIDGET_STATE (di));
    gdk_window_set_back_pixmap (widget->window, NULL, TRUE);

    if (di->is_floating)
        gdl_dock_item_window_float (di);
}

static void
gdl_dock_item_unrealize (GtkWidget *widget)
{
    GdlDockItem *di;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));

    di = GDL_DOCK_ITEM (widget);

    if (di->bin_window) {
        gdk_window_set_user_data (di->bin_window, NULL);
        gdk_window_destroy (di->bin_window);
        di->bin_window = NULL;
    };

    if (di->float_window) {
        gdk_window_set_user_data (di->float_window, NULL);
        gdk_window_destroy (di->float_window);
        di->float_window = NULL;
    };

    if (GTK_WIDGET_CLASS (parent_class)->unrealize)
        (* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
}


static void
gdl_dock_item_style_set (GtkWidget *widget,
                         GtkStyle  *previous_style)
{
    GdlDockItem *di;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));

    di = GDL_DOCK_ITEM (widget);

    if (GTK_WIDGET_REALIZED (widget) &&
        !GTK_WIDGET_NO_WINDOW (widget)) {
        gtk_style_set_background (widget->style, widget->window,
                                  widget->state);
        gtk_style_set_background (widget->style, di->bin_window, 
                                  widget->state);
        gtk_style_set_background (widget->style, di->float_window, 
                                  widget->state);
        if (GTK_WIDGET_DRAWABLE (widget))
            gdk_window_clear (widget->window);
    }
}

static void
draw_textured_frame (GtkWidget *widget, GdkWindow *window, 
                     GdkRectangle *rect, GtkShadowType shadow,
                     GdkRectangle *clip)
{
    /* FIXME: make a nicer dragbar */
    /* FIXME: if the item is floating and is resizable draw a handle
       to resize it and implement the event hooks */
    gtk_paint_handle (widget->style, window, GTK_STATE_NORMAL, shadow,
                      clip, widget, "dockitem",
                      rect->x, rect->y, rect->width, rect->height, 
                      GTK_ORIENTATION_VERTICAL);
}

static void
gdl_dock_item_paint (GtkWidget      *widget,
                     GdkEventExpose *event,
                     GdkRectangle   *area)
{
    GtkBin *bin;
    GdlDockItem *di;
    guint width;
    guint height;
    guint border_width;
    GdkRectangle rect;
    gint drag_handle_size = DRAG_HANDLE_SIZE;

    if (!GDL_DOCK_ITEM_NOT_LOCKED (widget))
        drag_handle_size = 0;

    bin = GTK_BIN (widget);
    di = GDL_DOCK_ITEM (widget);

    border_width = GTK_CONTAINER (di)->border_width;

    if (di->is_floating) {
        width = bin->child->allocation.width + 2 * border_width;
        height = bin->child->allocation.height + 2 * border_width;
    } else if (di->orientation == GTK_ORIENTATION_HORIZONTAL) {
        width = widget->allocation.width - drag_handle_size;
        height = widget->allocation.height;
    } else {
        width = widget->allocation.width;
        height = widget->allocation.height - drag_handle_size;
    }

    if (!event)
        gtk_paint_box (widget->style,
                       di->bin_window,
                       GTK_WIDGET_STATE (widget),
                       GTK_SHADOW_NONE,
                       area, widget,
                       "dockitem_bin",
                       0, 0, -1, -1);
    else
        gtk_paint_box (widget->style,
                       di->bin_window,
                       GTK_WIDGET_STATE (widget),
                       GTK_SHADOW_NONE,
                       &event->area, widget,
                       "dockitem_bin",
                       0, 0, -1, -1);

    /* We currently draw the handle _above_ the relief of the dockitem.
       It could also be drawn on the same level...  */
    if (GDL_DOCK_ITEM_NOT_LOCKED (widget)) {
        GdkRectangle dest;
    
        rect.x = 0;
        rect.y = 0;
      
        if (di->orientation == GTK_ORIENTATION_HORIZONTAL) {
            rect.width = DRAG_HANDLE_SIZE;
            rect.height = height;
        } else {
            rect.width = width;
            rect.height = DRAG_HANDLE_SIZE;
        }

        if (gdk_rectangle_intersect (event ? &event->area : area, 
                                     &rect, &dest))
            draw_textured_frame (widget, di->bin_window, &rect,
                                 GTK_SHADOW_OUT,
                                 event ? &event->area : area);
    }
  
    if (bin->child && GTK_WIDGET_VISIBLE (bin->child)) {
        GdkRectangle child_area;
        GdkEventExpose child_event;
        if (!event) { /* we were called from draw() */
            if (gtk_widget_intersect (bin->child, area, &child_area))
                gtk_widget_draw (bin->child, &child_area);
        } else { /* we were called from expose() */
            child_event = *event;
          
            if (GTK_WIDGET_NO_WINDOW (bin->child) &&
                gtk_widget_intersect (bin->child, &event->area, 
                                      &child_event.area))
                gtk_widget_event (bin->child, (GdkEvent *) &child_event);
        }
    }
}

static void
gdl_dock_item_draw (GtkWidget    *widget,
                    GdkRectangle *area)
{
    GdlDockItem *di;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GDL_IS_DOCK_ITEM (widget));

    di = GDL_DOCK_ITEM (widget);

    if (GTK_WIDGET_DRAWABLE (widget)) {
        if (di->is_floating) {
            GdkRectangle r;

            /* The area parameter does not make sense in this case, so
               we repaint everything.  */
            r.x = 0;
            r.y = 0;
            r.width = (2 * GTK_CONTAINER (di)->border_width
                       + DRAG_HANDLE_SIZE);
            r.height = r.width + GTK_BIN (di)->child->allocation.height;
            r.width += GTK_BIN (di)->child->allocation.width;

            gdl_dock_item_paint (widget, NULL, &r);

        } else
            gdl_dock_item_paint (widget, NULL, area);
    }
}

static gint
gdl_dock_item_expose (GtkWidget      *widget,
                      GdkEventExpose *event)
{
    g_return_val_if_fail (widget != NULL, FALSE);
    g_return_val_if_fail (GDL_IS_DOCK_ITEM (widget), FALSE);
    g_return_val_if_fail (event != NULL, FALSE);

    if (GTK_WIDGET_DRAWABLE (widget) && event->window != widget->window)
        gdl_dock_item_paint (widget, event, NULL);
  
    return FALSE;
}

static void
gdl_dock_item_set_floating (GdlDockItem *item, 
                            gboolean val)
{
    item->is_floating = val;

    /* If there is a child and it supports the 'is_floating' flag
     * set that too.
     */
    if (item->bin.child != NULL) {
        GtkArgInfo *info_p;
        gchar *error = gtk_object_arg_get_info 
            (GTK_OBJECT_TYPE (item->bin.child), "is_floating", &info_p);
        if (error)
            g_free (error);
        else
            gtk_object_set (GTK_OBJECT (item->bin.child), 
                            "is_floating", val, NULL);
    }
}

static gint
gdl_dock_item_button_changed (GtkWidget *widget,
                              GdkEventButton *event)
{
    GdlDockItem *di;
    gboolean event_handled;
  
    g_return_val_if_fail (widget != NULL, FALSE);
    g_return_val_if_fail (GDL_IS_DOCK_ITEM (widget), FALSE);
    g_return_val_if_fail (event != NULL, FALSE);

    di = GDL_DOCK_ITEM (widget);

    if (event->window != di->bin_window)
        return FALSE;

    if (!GDL_DOCK_ITEM_NOT_LOCKED (widget))
        return FALSE;

    event_handled = FALSE;

    if (event->button == 1 && event->type == GDK_BUTTON_PRESS) {
        GtkWidget *child;
        gboolean in_handle;
      
        child = GTK_BIN (di)->child;
      
        switch (di->orientation) {
	case GTK_ORIENTATION_HORIZONTAL:
            in_handle = event->x < DRAG_HANDLE_SIZE;
            break;
	case GTK_ORIENTATION_VERTICAL:
            in_handle = event->y < DRAG_HANDLE_SIZE;
            break;
	default:
            in_handle = FALSE;
            break;
	};

        if (!child) {
            in_handle = FALSE;
            event_handled = TRUE;
	};
      
        if (in_handle) {
            di->dragoff_x = event->x;
            di->dragoff_y = event->y;

            di->in_drag = TRUE;

            gdl_dock_item_grab_pointer (di);

            /* FIXME: should we use signals as GnomeDockItem does? */
            gdl_dock_drag_begin (GDL_DOCK (di->dock), di); 

            event_handled = TRUE;
	};
    } else if (event->type == GDK_BUTTON_RELEASE && di->in_drag) {
        gdk_pointer_ungrab (GDK_CURRENT_TIME);
      
        di->in_drag = FALSE;

        /* FIXME: should we use signals as GnomeDockItem does? */
        gdl_dock_drag_end (GDL_DOCK (di->dock), di);

        event_handled = TRUE;
    };

    return event_handled;
}

static gint
gdl_dock_item_motion (GtkWidget *widget,
                      GdkEventMotion *event)
{
    GdlDockItem *di;
    gint new_x, new_y;

    g_return_val_if_fail (widget != NULL, FALSE);
    g_return_val_if_fail (GDL_IS_DOCK_ITEM (widget), FALSE);
    g_return_val_if_fail (event != NULL, FALSE);

    di = GDL_DOCK_ITEM (widget);
    if (!di->in_drag)
        return FALSE;

    if (event->window != di->bin_window)
        return FALSE;

    gdk_window_get_pointer (NULL, &new_x, &new_y, NULL);
  
    /* FIXME: should we use signals as GnomeDockItem does? */
    gdl_dock_drag_motion (GDL_DOCK (di->dock), di, new_x, new_y);

    return TRUE;
}

static gint
gdl_dock_item_delete_event (GtkWidget *widget,
                            GdkEventAny *event)
{
    g_return_val_if_fail (widget != NULL, FALSE);
    g_return_val_if_fail (GDL_IS_DOCK_ITEM (widget), FALSE);
    g_return_val_if_fail (event != NULL, FALSE);

    return TRUE;
}

static void
gdl_dock_item_grab_pointer (GdlDockItem *item)
{
    GdkCursor *fleur;

    fleur = gdk_cursor_new (GDK_FLEUR);

    /* Hm, not sure this is the right thing to do, but it seems to work.  */
    while (gdk_pointer_grab (item->bin_window,
                             FALSE,
                             (GDK_BUTTON1_MOTION_MASK |
                              GDK_POINTER_MOTION_HINT_MASK |
                              GDK_BUTTON_RELEASE_MASK),
                             NULL,
                             fleur,
                             GDK_CURRENT_TIME) != 0);

    gdk_cursor_destroy (fleur);
}


/* ----------------------------------------------------------------------
 * Public interface 
 * ---------------------------------------------------------------------- */

GtkWidget *
gdl_dock_item_new (const gchar *name,
                   GdlDockItemBehavior behavior)
{
    GdlDockItem *dock_item;

    dock_item = GDL_DOCK_ITEM (gtk_type_new (gdl_dock_item_get_type ()));

    dock_item->name = g_strdup (name);
    dock_item->behavior = behavior;

    return GTK_WIDGET (dock_item);
}

guint
gdl_dock_item_get_type (void)
{
    static GtkType dock_item_type = 0;

    if (dock_item_type == 0) {
        GtkTypeInfo dock_item_info = {
            "GdlDockItem",
            sizeof (GdlDockItem),
            sizeof (GdlDockItemClass),
            (GtkClassInitFunc) gdl_dock_item_class_init,
            (GtkObjectInitFunc) gdl_dock_item_init,
            /* reserved_1 */ NULL,
            /* reserved_2 */ NULL,
            (GtkClassInitFunc) NULL,
        };

        dock_item_type = gtk_type_unique (gtk_bin_get_type (), 
                                          &dock_item_info);
    }

    return dock_item_type;
}

void
gdl_dock_item_dock_to (GdlDockItem *item,
                       GdlDockItem *target_item,
                       GdlDockPlacement position)
{
    GtkWidget *widget;
    GtkWidget *old_parent;
    gboolean auto_reduce_parent = TRUE;

    g_return_if_fail (item != NULL);
    g_return_if_fail (item != target_item);
    g_return_if_fail (target_item != NULL || position == GDL_DOCK_FLOATING);

    DEBUG (("entering function"));

    widget = GTK_WIDGET (item);
    gtk_widget_ref (widget);

    old_parent = widget->parent;

    if (position == GDL_DOCK_FLOATING || target_item == NULL) {
        if (old_parent != NULL)
            gtk_container_remove (GTK_CONTAINER (old_parent), widget);

        if (!item->is_floating) {
            gdl_dock_add_floating_item (GDL_DOCK (item->dock), item, 
                                        item->float_x, item->float_y, 
                                        item->orientation);
        };

    } else {
        GtkWidget *target, *parent;

        target = GTK_WIDGET (target_item);
        parent = target->parent;

        /* propagate dock object */
        item->dock = target_item->dock;

        /* if the widget is in the same container as the target */
        if (parent == old_parent && GDL_IS_DOCK_PANED (parent)) {
            /* FIXME: maybe this should be a virtual of GdlDockItem */
            gdl_dock_paned_reorder (GDL_DOCK_PANED (parent), widget, position);
            auto_reduce_parent = FALSE;

        } else {
            GdlDockPaned *paned;

            if (old_parent != NULL)
                gtk_container_remove (GTK_CONTAINER (old_parent), widget);

            /* general case */
            gtk_widget_ref (target);
            gtk_container_remove (GTK_CONTAINER (parent), target);

            if (position == GDL_DOCK_LEFT || position == GDL_DOCK_RIGHT) {
                paned = GDL_DOCK_PANED (gdl_dock_paned_new 
                                        (GTK_ORIENTATION_HORIZONTAL));
            } else {
                paned = GDL_DOCK_PANED (gdl_dock_paned_new 
                                        (GTK_ORIENTATION_VERTICAL));
            };

            /* GdkWindow stuff */
            if (item->is_floating)
                gdl_dock_item_window_sink (item);

            if (position == GDL_DOCK_LEFT || position == GDL_DOCK_TOP) {
                gdl_dock_paned_add1 (paned, widget);
                gdl_dock_paned_add2 (paned, target);
            } else {
                gdl_dock_paned_add1 (paned, target);
                gdl_dock_paned_add2 (paned, widget);
            };
            gtk_widget_unref (target);

            gtk_widget_show (GTK_WIDGET (paned));
            gtk_container_add (GTK_CONTAINER (parent), GTK_WIDGET (paned));
        };
    };

    if (old_parent != NULL && auto_reduce_parent) {
        if (GDL_IS_DOCK_ITEM (old_parent)) {
            gdl_dock_item_auto_reduce (GDL_DOCK_ITEM (old_parent));
        };
    };
    gtk_widget_unref (widget);
}

/* FIXME: make this virtual so the paned can change accordingly */
gboolean
gdl_dock_item_set_orientation (GdlDockItem *dock_item,
                               GtkOrientation orientation)
{
    GtkArgInfo *info_p;
    gchar *error;

    g_return_val_if_fail (dock_item != NULL, FALSE);
    g_return_val_if_fail (GDL_IS_DOCK_ITEM (dock_item), FALSE);

    if (dock_item->orientation != orientation) {
        dock_item->orientation = orientation;

        if (dock_item->bin.child != NULL) {
            error = gtk_object_arg_get_info 
                (GTK_OBJECT_TYPE (dock_item->bin.child),
                 "orientation", &info_p);
            if (error)
                g_free (error);
            else {
                if (orientation == GTK_ORIENTATION_HORIZONTAL) {
                    gtk_object_set (GTK_OBJECT (dock_item->bin.child),
                                    "orientation", GTK_ORIENTATION_VERTICAL,
                                    NULL);
                } else {
                    gtk_object_set (GTK_OBJECT (dock_item->bin.child),
                                    "orientation", GTK_ORIENTATION_HORIZONTAL,
                                    NULL);
                };
            };
        }
        if (GTK_WIDGET_DRAWABLE (dock_item))
            gtk_widget_queue_clear (GTK_WIDGET (dock_item));
        gtk_widget_queue_resize (GTK_WIDGET (dock_item));
    }

    return TRUE;
}

void
gdl_dock_item_auto_reduce (GdlDockItem *item)
{
    GdlDockItemClass *klass = GDL_DOCK_ITEM_CLASS (GTK_OBJECT (item)->klass);

    if (klass->auto_reduce)
        (* klass->auto_reduce)(item);
}


#define SPLIT_RATIO  0.3

gboolean
gdl_dock_item_drag_request (GdlDockItem *item, 
                            gint x,
                            gint y, 
                            GdlDockRequestInfo *target)
{
    GdlDockItemClass *klass = GDL_DOCK_ITEM_CLASS (GTK_OBJECT (item)->klass);

    if (klass->drag_request)
        return (* klass->drag_request) (item, x, y, target);
    else {
        GtkAllocation *alloc;
        gint dx, dy;

        alloc = &(GTK_WIDGET (item)->allocation);

        dx = x - alloc->x;
        dy = y - alloc->y;

        if (dx > 0 && dx < alloc->width && dy > 0 && dy < alloc->height) {
            /* it's inside the item */
            float rx, ry;
            rx = (float) dx / alloc->width;
            ry = (float) dy / alloc->height;

            if (rx < 0.5)
                target->position = GDL_DOCK_LEFT;
            else {
                target->position = GDL_DOCK_RIGHT;
                rx = 1 - rx;
            };
            
            if (ry < 0.5 && ry < rx)
                target->position = GDL_DOCK_TOP;
            else if (ry > 0.5 && (1 - ry) < rx)
                target->position = GDL_DOCK_BOTTOM;

            target->rect.x = 0;
            target->rect.y = 0;
            target->rect.width = alloc->width;
            target->rect.height = alloc->height;

            if (target->requestor != item) {
                switch (target->position) {
                case GDL_DOCK_TOP:
                    target->rect.height *= SPLIT_RATIO;
                    break;
                case GDL_DOCK_BOTTOM:
                    target->rect.y += target->rect.height * (1 - SPLIT_RATIO);
                    target->rect.height *= SPLIT_RATIO;
                    break;
                case GDL_DOCK_LEFT:
                    target->rect.width *= SPLIT_RATIO;
                    break;
                case GDL_DOCK_RIGHT:
                    target->rect.x += target->rect.width * (1 - SPLIT_RATIO);
                    target->rect.width *= SPLIT_RATIO;
                    break;
                default:
                    break;
                };
            };
            
            target->target = GTK_WIDGET (item);

            return TRUE;
                
        } else
            return FALSE;
    };
}

void
gdl_dock_item_drag_floating (GdlDockItem *item, 
                             gint x, 
                             gint y)
{
    x -= item->dragoff_x;
    y -= item->dragoff_y;

    if (item->is_floating) {
        gdk_window_move (item->float_window, x, y);
        gdk_window_raise (item->float_window);
    };
    item->float_x = x;
    item->float_y = y;
}

void
gdl_dock_item_window_sink (GdlDockItem *item)
{
    if (GTK_WIDGET_REALIZED (GTK_WIDGET (item))) {
        gdk_window_move_resize (GTK_WIDGET (item)->window, -1, -1, 0, 0);
        gdk_window_hide (item->float_window);
        gdk_window_reparent (item->bin_window, 
                             GTK_WIDGET (item)->window, 0, 0);
        gdk_window_show (GTK_WIDGET (item)->window);
    };

    item->float_window_mapped = FALSE;
    gdl_dock_item_set_floating (item, FALSE);
}

void
gdl_dock_item_window_float (GdlDockItem *item)
{
    GtkRequisition requisition;
    GtkAllocation allocation;

    item->is_floating = TRUE;
    item->dock = GTK_WIDGET (item)->parent;

    if (! GTK_WIDGET_REALIZED (item))
        return;

    gtk_widget_size_request (GTK_WIDGET (item), &requisition);

    gdk_window_move_resize (item->float_window,
                            item->float_x, item->float_y, 
                            requisition.width, requisition.height);

    gdk_window_reparent (item->bin_window, item->float_window, 0, 0);
    gdk_window_set_hints (item->float_window, item->float_x, item->float_y, 
                          0, 0, 0, 0, GDK_HINT_POS);

    gdk_window_show (item->float_window);
    item->float_window_mapped = TRUE;

    allocation.x = allocation.y = 0;
    allocation.width = requisition.width;
    allocation.height = requisition.height;
    gtk_widget_size_allocate (GTK_WIDGET (item), &allocation);

    gdk_window_hide (GTK_WIDGET (item)->window);
    gdl_dock_item_draw (GTK_WIDGET (item), NULL);

    gdk_window_set_transient_for
        (item->float_window, gdk_window_get_toplevel
         (GTK_WIDGET (item)->window));
}

