/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#ifndef __GDL_DOCK_ITEM_H__
#define __GDL_DOCK_ITEM_H__

#include <gtk/gtk.h>


#define GDL_TYPE_DOCK_ITEM            (gdl_dock_item_get_type ())
#define GDL_DOCK_ITEM(obj)            (GTK_CHECK_CAST ((obj), GDL_TYPE_DOCK_ITEM, GdlDockItem))
#define GDL_DOCK_ITEM_CLASS(klass)    (GTK_CHECK_CLASS_CAST ((klass), GDL_TYPE_DOCK_ITEM, GdlDockItemClass))
#define GDL_IS_DOCK_ITEM(obj)         (GTK_CHECK_TYPE ((obj), GDL_TYPE_DOCK_ITEM))
#define GDL_IS_DOCK_ITEM_CLASS(klass) (GTK_CHECK_CLASS_TYPE ((klass), GDL_TYPE_DOCK_ITEM))


typedef enum {
    GDL_DOCK_TOP,
    GDL_DOCK_BOTTOM,
    GDL_DOCK_RIGHT,
    GDL_DOCK_LEFT,
    GDL_DOCK_FLOATING
} GdlDockPlacement;

typedef enum {
    GDL_DOCK_ITEM_BEH_NORMAL = 0,
    GDL_DOCK_ITEM_BEH_EXCLUSIVE = 1 << 0,
    GDL_DOCK_ITEM_BEH_NEVER_FLOATING = 1 << 1,
    GDL_DOCK_ITEM_BEH_NEVER_VERTICAL = 1 << 2,
    GDL_DOCK_ITEM_BEH_NEVER_HORIZONTAL = 1 << 3,
    GDL_DOCK_ITEM_BEH_LOCKED = 1 << 4
} GdlDockItemBehavior;


typedef struct _GdlDockItem        GdlDockItem;
typedef struct _GdlDockItemClass   GdlDockItemClass;
typedef struct _GdlDockRequestInfo GdlDockRequestInfo;

struct _GdlDockItem {
    GtkBin bin;

    GdkWindow           *bin_window;
    GdkWindow           *float_window;

    gchar               *name;
    GdlDockItemBehavior  behavior;
    GtkOrientation       orientation;
    GtkWidget           *dock;

    guint      is_floating : 1;
    guint      float_window_mapped : 1;
    guint      in_drag : 1;
    guint      resize : 1;
    guint      shrink : 1;
    gint       dragoff_x, dragoff_y;

    gint       float_x, float_y;
};

/* structure for drag_request return information */
struct _GdlDockRequestInfo {
    /* target is a GtkWidget because the dock can be a target */
    GdlDockItem *requestor;
    GtkWidget *target;
    GdlDockPlacement position;
    GdkRectangle rect;
};

struct _GdlDockItemClass {
    GtkBinClass parent_class;

    /* virtuals */
    void (* auto_reduce)      (GdlDockItem *item);
    gboolean (* drag_request) (GdlDockItem *item, 
                               gint x, gint y, 
                               GdlDockRequestInfo *target);
};



GtkWidget     *gdl_dock_item_new               (const gchar *name,
                                                GdlDockItemBehavior behavior);

guint          gdl_dock_item_get_type          (void);

void           gdl_dock_item_dock_to           (GdlDockItem *item,
                                                GdlDockItem *target,
                                                GdlDockPlacement position);

/* FIXME: make it virtual or private */
gboolean       gdl_dock_item_set_orientation   (GdlDockItem *item,
                                                GtkOrientation orientation);

void           gdl_dock_item_auto_reduce       (GdlDockItem *item);

gboolean       gdl_dock_item_drag_request      (GdlDockItem *item, 
                                                gint x,
                                                gint y, 
                                                GdlDockRequestInfo *target);

/* GdkWindow setup and stuff */
void           gdl_dock_item_window_sink       (GdlDockItem *item);
void           gdl_dock_item_window_float      (GdlDockItem *item);
void           gdl_dock_item_drag_floating     (GdlDockItem *item, 
                                                gint x, 
                                                gint y);



#endif
